/* Public domain. Written by Stone Fox */
#include <stdio.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "cgi.h"

#if CGIDEBUG
#define DEBUG_PUTSTR(str) putstr(2, str);
#define DEBUG_PUT_LINE(str) put_line(2, str);
#define DEBUG_NEW_LINE(count) new_line(2, count);
#else
#define DEBUG_PUTSTR(str)
#define DEBUG_PUTINT(value)
#define DEBUG_PUT_LINE(str)
#define DEBUG_NEW_LINE(count)
#endif

struct form_entry {
	char *attr;
	char *value;
	int value_length;
	struct form_entry *next;
};

char *request_method;
char *query_string;
char content_type_data[1024];
char *content_type = &content_type_data[0];
int content_length;
int hex_value[256];
int num_of_form_entries = 0;
struct form_entry *form_entry_first;
static char *cgi_find_target = 0;
static struct form_entry *cgi_find_pos = 0;

int tolower(int ch)
{
	if (ch - 'A' < 26)
		return ch | 32;
	return ch;
}

static int isalpha(int ch)
{
	return (ch | 32) - 'a' < 26;
}

static int isdigit(unsigned int ch)
{
	return ch - '0' < 10;
}

static void cgi_free_resources(void)
{
	int len = 0;
	struct form_entry *c = form_entry_first;
	struct form_entry *n;

	while (c) {
		n = (*c).next;
		c = n;
		len++;
	}
	sbrk(-(content_length + sizeof(struct form_entry) * len));
}

static void setup_constants(void)
{
	memset(hex_value, 0, sizeof(hex_value));
	hex_value['0'] = 0;
	hex_value['1'] = 1;
	hex_value['2'] = 2;
	hex_value['3'] = 3;
	hex_value['4'] = 4;
	hex_value['5'] = 5;
	hex_value['6'] = 6;
	hex_value['7'] = 7;
	hex_value['8'] = 8;
	hex_value['9'] = 9;
	hex_value['A'] = 10;
	hex_value['B'] = 11;
	hex_value['C'] = 12;
	hex_value['D'] = 13;
	hex_value['E'] = 14;
	hex_value['F'] = 15;
	hex_value['a'] = 10;
	hex_value['b'] = 11;
	hex_value['c'] = 12;
	hex_value['d'] = 13;
	hex_value['e'] = 14;
	hex_value['f'] = 15;
}

static int str_eq_nc(char *s1, char *s2)
{
	while (1) {
		if (!(*s1)) {
			if (!(*s2)) {
				return 1;
			} else {
				return 0;
			}
		} else if (!(*s2)) {
			return 0;
		}
		if (isalpha(*s1)) {
			if (tolower(*s1) != tolower(*s2))
				return 0;
		} else if ((*s1) != (*s2)) {
			return 0;
		}
		s1++;
		s2++;
	}
}

static int str_begins_nc(char *s1, char *s2)
{
	while(1) {
		if (!(*s2)) {
			return 1;
		} else if (!(*s1)) {
			return 0;
		}
		if (isalpha(*s1)) {
			if (tolower(*s1) != tolower(*s2)) {
				return 0;
			}
		} else if ((*s1) != (*s2)) {
			return 0;
		}
		s1++;
		s2++;
	}
}

void cgi_header_location(char *redirect_url)
{
	put(1, "Location:");
	put(1, redirect_url);
	put(1, "\r\n\r\n");
}

void cgi_header_status(int status, char *status_message)
{
	char buffer[4];

	put(1, "Status:");
	itoa(status, buffer, 10);
	put(1, buffer);
	put(1, " ");
	put(1, status_message);
	put(1, "\r\n\r\n");
}

void cgi_header_content_type(char *mime_type)
{
	put(1, "Content-type:");
	put(1, mime_type);
	put(1, "\r\n\r\n");
}

int cgi_unescape_chars(char *s, char *cp, int len)
{
	enum {
		escape_rest,
		escape_first,
		escape_second
	};
	int escape_state = escape_rest;
	int escaped_value = 0;
	int src_pos = 0;
	int dst_pos = 0;
	int ch = cp[src_pos];

	while (src_pos < len) {
		switch (escape_state) {
		case escape_rest:
			if (ch == '%') {
				escape_state = escape_first;
			} else if (ch == '+') {
				s[dst_pos++] = ' ';
			} else {
				s[dst_pos++] = ch;	
			}
			break;
		case escape_first:
			escaped_value = hex_value[ch] << 4;	
			escape_state = escape_second;
			break;
		case escape_second:
			escaped_value += hex_value[ch];
			s[dst_pos++] = escaped_value;
			escape_state = escape_rest;
			break;
		}
		src_pos++;
	}
	s[dst_pos] = '\0';
	return dst_pos;
}

/* Scan for pairs, unescaping and storing them as they are found. */
static int parse_form_input(char *data, int length)
{
	int pos = 0;
	struct form_entry *n;
	struct form_entry *l = 0;
	int found_amp = 0;
	int start = pos;
	int len = 0;

	while (pos != length) {
		while (pos != length) {
			if (data[pos] == '&')
				break;
			if (data[pos] == '=') {
				pos++;
				break;
			}
			pos++;
			len++;
		}
		if (!len)
			break;
		n = sbrk(sizeof(struct form_entry));
		if (!n)
			return 1;

		/* data is overwritten by (*n).attr and (*n).value as its parsed, so no
                 * additional memory allocation is needed
                 */
		(*n).attr = data + start;
		cgi_unescape_chars((*n).attr, (*n).attr, len);

		start = pos;
		len = 0;
		while (pos != length) {
			if (data[pos] == '&') {
				found_amp = 1;
				pos++;
				break;
			}
			pos++;
			len++;
		}
		(*n).value = data + start;
		(*n).value_length = cgi_unescape_chars((*n).value, (*n).value, len);
		(*n).next = 0;
		if (!l) {
			form_entry_first = n;
		} else {
			(*l).next = n;
		}
		l = n;
		if (!found_amp)
			break;
	}
	return 0;
}

static int parse_post_form_input(void)
{
	char *input = sbrk(content_length);
	int result;

	if (!content_length)
		return 0;
	if (read(0, input, content_length) != content_length) 
		return 1;
	result = parse_form_input(input, content_length);
	return result;
}

int main(int argc, char **argv, char **envp)
{
	int result;
	char *content_length_string;
	char *e;

	setup_constants();
	request_method = getenv(envp, "REQUEST_METHOD");
	query_string = getenv(envp, "QUERY_STRING");
	e = getenv(envp, "CONTENT_TYPE");
	if (e) {
		result = strlen(e);
		if (result < sizeof(content_type_data)) {
			memcpy(content_type, e, result + 1);
		} else {
			memcpy(content_type, e, sizeof(content_type_data));
			content_type[sizeof(content_type_data) - 1] = '\0';
		}
	} else {
		content_type[0] = '\0';
	}

	content_length_string = getenv(envp, "CONTENT_LENGTH");
	content_length = strtoi(content_length_string, NULL, 10);
	DEBUG_PUTINT(content_length);
	DEBUG_NEW_LINE(1);
	DEBUG_PUT_LINE(request_method);
	DEBUG_PUT_LINE(content_type);
	form_entry_first = 0;

	if (str_eq_nc(request_method, "post")) {
		DEBUG_PUT_LINE("POST recognized");
		if (str_eq_nc(content_type, "application/x-www-form-urlencoded")) {	
			DEBUG_PUT_LINE("Calling post_form_input");
			if (parse_post_form_input() != 0) {
				DEBUG_PUT_LINE("post_form_input failed");
				cgi_header_status(500, "Error reading form data");
				form_entry_first = 0;
				return 1;
			}	
			DEBUG_PUT_LINE("post_form_input succeeded");
		} else if (str_eq_nc(content_type, "multipart/form-data")) {
			DEBUG_PUT_LINE("multipart/form-data not implemented");
			cgi_header_status(500, "Error reading form data");
			form_entry_first = 0;
			return 1;
		}
	} else if (str_eq_nc(request_method, "get")) {
		/* The spec says this should be taken care of by the server, but it isn't */
		content_length = strlen(query_string);
		if (parse_form_input(query_string, content_length) != 0) {
			DEBUG_PUT_LINE("get_form_input failed");
			cgi_header_status(500, "Error reading form data");
			form_entry_first = 0;
			return 1;
		} else {
			DEBUG_PUT_LINE("get_form_input succeeded");
		}
	}
	result = cgi_main(argc, argv, envp);
	cgi_free_resources();
	return result;
}

static int form_entry_string(struct form_entry *e, char *result, int max, int newlines)
{
	char *dp, *sp;
	int truncated = 0;
	int len = 0;
	int avail = max-1;
	int cr_count = 0;
	int lf_count = 0;
	int ch;
	int lfs_add;

	dp = result;
	sp = (*e).value;	

	while (1) {
		ch = *sp;
		if (ch == 13 || ch == 10) {
			if (ch == 13) {
				cr_count++;
			} else {
				lf_count++;
			}
		} else {
			if (cr_count || lf_count) {
				lfs_add = cr_count;
				if (lf_count > cr_count)
					lfs_add = lf_count;
				/* Stomp all newlines if desired */
				if (!newlines)
					lfs_add = 0;
				while (lfs_add) {
					if (len >= avail) {
						truncated = 1;
						break;
					}
					*dp = 10;
					dp++;
					lfs_add--;
					len++;		
				}
				cr_count = 0;
				lf_count = 0;
			}
			if (ch == '\0')
				break;				
			if (len >= avail) {
				truncated = 1;
				break;
			}
			*dp = ch;
			dp++;
			len++;
		}
		sp++;	
	}	
	*dp = '\0';
	if (truncated)
		return form_truncated;
	if (!len)
		return form_empty;
	return form_success;
}

static struct form_entry *form_entry_find_next(void)
{
	struct form_entry *c;

	while (cgi_find_pos) {
		c = cgi_find_pos;
		cgi_find_pos = (*c).next;
		if (!strcmp((*c).attr, cgi_find_target))
			return c;
	}
	return 0;
}

static struct form_entry *form_entry_find_first(char *name)
{
	cgi_find_target = name;
	cgi_find_pos = form_entry_first;
	return form_entry_find_next();
}

int cgi_form_string(char *name, char *result, int max)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		*result = '\0';
		return form_not_found;
	}
	return form_entry_string(e, result, max, 1);
}

int cgi_form_string_no_newlines(char *name, char *result, int max)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		*result = '\0';
		return form_not_found;
	}
	return form_entry_string(e, result, max, 0);
}

void cgi_string_array_free(char **string_array)
{
}

int cgi_form_string_multiple(char *name, char ***result)
{
	char *string_array[64];
	struct form_entry *e;
	int i;
	int total = 0;
	int max;

	/* TODO: limit number of entries to size of string_array */
	e = form_entry_find_first(name);
	if (e != 0) {
		do {
			total++;
		} while ((e = form_entry_find_next()) != 0); 
	}
	/* initialize all entries to null; the last will stay that way */
	for (i = 0; i <= total; i++)
		string_array[i] = 0;
	/* Now go get the entries */
	e = form_entry_find_first(name);
	DEBUG_PUT_LINE("string_multiple Beginning");
	if (e) {
		i = 0;
		do {
			max = (*e).value_length + 1;
			string_array[i] = sbrk(max);
			if (string_array[i] == 0) {
				/* Memory problems */
				cgi_string_array_free(string_array);
				*result = 0;
				return form_memory;
			}
			memcpy(string_array[i], (*e).value, max);
			form_entry_string(e, string_array[i], max, 1);
			i++;
		} while ((e = form_entry_find_next()) != 0); 
		*result = string_array;
		DEBUG_PUT_LINE("string_multiple Succeeding");
		return form_success;
	} else {
		*result = string_array;
		DEBUG_PUT_LINE("string_multiple found nothing");
		return form_not_found;
	}
}

int cgi_form_string_space_needed(char *name, int *result)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		*result = 1;
		return form_not_found; 
	}
	*result = strlen((*e).value) + 1;
	return form_success;
}

static int cgi_first_nonspace_char(char *s)
{
	int len = strspn(s, " \n\r\t");
	return s[len];
}

int cgi_form_integer(char *name, int *result, int default_v)
{
	struct form_entry *e;
	int ch;

	e = form_entry_find_first(name);
	if (!e) {
		*result = default_v;
		return form_not_found; 
	}	
	if (!strlen((*e).value)) {
		*result = default_v;
		return form_empty;
	}
	ch = cgi_first_nonspace_char((*e).value);
	if (!(isdigit(ch)) && (ch != '-') && (ch != '+')) {
		*result = default_v;
		return form_bad_type;
	}
	*result = strtoi((*e).value, NULL, 10);
	return form_success;
}

int cgi_form_integer_bounded(char *name, int *result, int min, int max, int default_v)
{
	int error = cgi_form_integer(name, result, default_v);

	if (error != form_success)
		return error;
	if (*result < min) {
		*result = min;
		return form_constrained;
	} 
	if (*result > max) {
		*result = max;
		return form_constrained;
	} 
	return form_success;
}

int cgi_form_select_single(char *name, char **choices_text, int choices_total,
int *result, int default_v)
{
	struct form_entry *e;
	int i;

	e = form_entry_find_first(name);
	DEBUG_PUTINT((int)e);
	DEBUG_NEW_LINE(1);
	if (!e) {
		*result = default_v;
		return form_not_found;
	}
	for (i = 0; i < choices_total; i++) {
		DEBUG_PUTSTR(choices_text[i]);
		DEBUG_PUTSTR(" ");
		DEBUG_PUT_LINE((*e).value);
		if (!strcmp(choices_text[i], (*e).value)) {
			DEBUG_PUT_LINE("MATCH");
			*result = i;
			return form_success;
		}
	}
	*result = default_v;
	return form_no_such_choice;
}

int cgi_form_select_multiple(char *name, char **choices_text, int choices_total,
int *result, int *invalid)
{
	struct form_entry *e;
	int i;
	int hits = 0;
	int invalid_e = 0;
	int hit;

	for (i = 0; i < choices_total; i++)
		result[i] = 0;
	e = form_entry_find_first(name);
	if (!e) {
		*invalid = invalid_e;
		return form_not_found;
	}
	do {
		hit = 0;
		for (i = 0; i < choices_total; i++) {
			if (!strcmp(choices_text[i], (*e).value)) {
				result[i] = 1;
				hits++;
				hit = 1;
				break;
			}
		}
		if (!(hit)) {
			invalid_e++;
		}
	} while ((e = form_entry_find_next()) != 0);

	*invalid = invalid_e;

	if (hits)
		return form_success;
	return form_not_found;
}

int cgi_form_checkbox_single(char *name)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e)
		return form_not_found;
	return form_success;
}

int cgi_form_entries(char ***result)
{
	char *string_array[64];
	struct form_entry *e, *pe;
	int i;
	int total = 0;
	unsigned int space;

	e = form_entry_first;
	while (e && total < sizeof(*string_array)) {
		/* Don't count a field name more than once if multiple values happen to
		 * be present for it
		 */
		pe = form_entry_first;
		while (pe != e) {
			if (!strcmp((*e).attr, (*pe).attr)) {
				goto skip_second_value;
			}
			pe = (*pe).next;					
		}
		total++;
skip_second_value:
		e = (*e).next;
	}
	/* initialize all entries to null; the last will stay that way */
	for (i = 0; i <= total; i++)
		string_array[i] = 0;
	/* Now go get the entries */
	e = form_entry_first;
	i = 0;
	while (e) {
		/* Don't return a field name more than once if
			multiple values happen to be present for it */
		pe = form_entry_first;
		while (pe != e) {
			if (!strcmp((*e).attr, (*pe).attr))
				goto skip_second_value2;
			pe = (*pe).next;					
		}		
		space = strlen((*e).attr) + 1;
		string_array[i] = sbrk(space);
		if (string_array[i] == 0) {
			cgi_string_array_free(string_array);
			*result = 0;
			return 1;
		}
		memcpy(string_array[i], (*e).attr, space);
		i++;
skip_second_value2:
		e = (*e).next;
	}
	*result = string_array;
	return form_success;
}

int cgi_xhtml_escape(const char *data, unsigned int len)
{
	while (len--) {
		if (*data == '<') {
			put(1, "&lt;");
		} else if (*data == '&') {
			put(1, "&amp;");
		} else if (*data == '>') {
			put(1, "&gt;");
		} else {
			write(1, data, 1);
		}
		data++;
	}
	return form_success;
}

int cgi_value_escape(const char *data, unsigned int len)
{
	while (len--) {
		if (*data == '\"') {
			put(1, "&#34;");
		} else {
			write(1, data, 1);
		}
		data++;
	}
	return form_success;
}
