/* Make sure this is only included once. */
#ifndef CGI
#define CGI 1

/* Possible return codes from the Form family of functions (see below). */
enum {
	form_success,
	form_truncated,
	form_bad_type,
	form_empty,
	form_not_found,
	form_constrained,
	form_no_such_choice,
	form_memory,
	form_no_file_name,
	form_no_content_type,
	form_not_a_file,
	form_open_failed,
	form_IO,
	form_EOF
};

/* The number of bytes of data received. Note that if the submission is a
 * form submission the library will read and parse all the information
 * directly from standard input; the programmer need not do so.
 */
extern int cgi_content_length;

/* These functions are used to retrieve form data. See documents/function.txt for
 * documentation.
 */
int cgi_form_string(char *name, char *result, int max);
int cgi_form_string_no_newlines(char *name, char *result, int max);
int cgi_form_string_space_needed(char *name, int *length);
int cgi_form_string_multiple(char *name, char ***ptr_to_string_array);
void cgi_string_array_free(char **string_array);

int cgi_form_integer(char *name, int *result, int default_v);
int cgi_form_integer_bounded(char *name, int *result, int min, int max, int default_v);
int cgi_form_select_single(char *name, char **choices_text, int choices_total,
	int *result, int default_v);
int cgi_form_select_multiple(char *name, char **choices_text, int choices_total,
	int *result, int *invalid);
int cgi_form_checkbox_single(char *name);

/* path can be null or empty in which case a path of / (entire site) is set. name can be a
 * single web site; if it is an entire domain, such as 'boutell.dev', it should begin with
 * a dot: '.boutell.dev'
 */
void cgi_header_location(char *redirect_url);
void cgi_header_status(int status, char *status_message);
void cgi_header_content_type(char *mime_type);
int cgi_main(int argc, char **argv, char **envp);
int cgi_form_entries(char ***ptr_to_string_array);

/* Output data with the <, &, and > characters HTML-escaped. 'data' does not need to be
 * null-terminated; 'len' is the number of bytes in 'data'. Returns cgi_form_IO in the
 * event of error, cgi_form_success otherwise.
 */
int cgi_xhtml_escape(const char *data, unsigned int len);

/* Output data with the " character XHTML-escaped, and no other characters escaped. This is
 * useful when outputting the contents of a tag attribute such as 'href' or 'src'. 'data'
 * does not need to be null-terminated; 'len' is the number of bytes in 'data'. Returns
 * cgi_form_IO in the event of error, cgi_form_success otherwise.
 */
int cgi_value_escape(const char *data, unsigned int len);

#endif
