#include <stdio.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "cgi.h"

#if CGIDEBUG
#define DEBUG_PUTSTR(str) putstr(2, str);
#define DEBUG_PUT_LINE(str) put_line(2, str);
#define DEBUG_NEW_LINE(count) new_line(2, count);
#else
#define DEBUG_PUTSTR(str)
#define DEBUG_PUTINT(value)
#define DEBUG_PUT_LINE(str)
#define DEBUG_NEW_LINE(count)
#endif

#define TEMP_DIR "/tmp"
#define MAX_TEMP_SIZE 1073741824

struct form_entry {
	char *attr;
	char *value;
	int value_length;
	char *file_name;
	char *content_type;
	int t_file;
	struct form_entry *next;
};

struct mp_stream {
	char putback[1024];	
	int read_pos;
	int write_pos;
	int offset;
};

char *request_method;
char *query_string;
char content_type_data[1024];
char *content_type = &content_type_data[0];
char *multipart_boundary;
int content_length;
int hex_value[256];
int num_of_form_entries = 0;
struct form_entry *form_entry_first;
int stack_ptr;
unsigned chat memory[32768];

static int after_next_boundary(struct *mpStream mpp, int outf, char **out_p,
int *body_length_p, int first);
static int read_header_line(struct *mp_stream mpp, char *attr, int attr_space, char *value,
int value_space);
static void decompose_value(char *value, char *mvalue, int mvalue_space, char **arg_names,
char **arg_values, int arg_value_space);
static int get_temp_file(int *t_file);
static int cgi_unescape_chars(char **sp, char *cp, int len);
static int form_entry_string(struct form_entry *e, char *result, int max, int newlines);
static struct form_entry *form_entry_find_first(char *name);
static form_entry *form_entry_find_next();
static int cgi_first_nonspace_char(char *s);
static char *cgi_find_target = 0;
static struct form_entry *cgi_find_pos = 0;
static int parse_post_form_input(void);
static int parse_post_multipart_input(void);
static int parse_form_input(char *data, int length);
static void setup_constants(void);
static int str_eq_nc(char *s1, char *s2);
static int str_begins_nc(char *s1, char *s2);

#ifdef UNIT_TEST
static int unit_test(void);
#endif

int isspace(unsigned int ch)
{
	return ch == ' ' || ch - '\t' < 5;
}

void *allocate(unsigned int amount)
{
	if ((stack_len + amount) >= sizeof(memory))
		return NULL;
	return &memory[stack_len + amount];
}

int main(int argc, char **argv, char **envp)
{
	int result;
	char *content_length_string;
	char *e;
	char *sat;

	setup_constants();
	request_method = getenv(envp, "REQUEST_METHOD");
	query_string = getenv(envp, "QUERY_STRING");
	e = getenv(envp, "CONTENT_TYPE");
	if (e) {
		if (strlen(e) < sizeof(content_type_data)) {
			strcpy(content_type, e);
		} else {
			strncpy(content_type, e, sizeof(content_type_data));
			content_type[sizeof(content_type_data) - 1] = '\0';
		}
	} else {
		content_type[0] = '\0';
	}
	multipart_boundary = "";

	if (strchr(content_type, ';')) {
		sat = strchr(content_type, ';');
		while (sat) {
			*sat = '\0';
			sat++;
			while (isspace(*sat))
				sat++;
			if (str_begins_nc(sat, "boundary=")) {
				char *s;
				multipart_boundary = sat + strlen("boundary=");
				s = multipart_boundary;
				while ((*s) && (!isspace(*s)))
					s++;
				*s = '\0';
				break;
			} else {
				sat = strchr(sat, ';');
			} 	
		}
	}
	content_length_string = getenv(envp, "CONTENT_LENGTH");
	content_length = atoi(content_length_string);	
	DEBUG_PUTINT(content_length);
	DEBUG_NEW_LINE(1);
	DEBUG_PUT_LINE(request_method);
	DEBUG_PUT_LINE(content_type);
	form_entry_first = 0;

	if (str_eq_nc(request_method, "post")) {
		DEBUG_PUT_LINE("POST recognized");
		if (str_eq_nc(content_type, "application/x-www-form-urlencoded")) {	
			DEBUG_PUT_LINE("Calling post_form_input");
			if (parse_post_form_input() != 0) {
				DEBUG_PUT_LINE("post_form_input failed");
				cgi_header_status(500, "Error reading form data");
				form_entry_first = 0;
				return 1;
			}	
			DEBUG_PUT_LINE("post_form_input succeeded");
		} else if (str_eq_nc(content_type, "multipart/form-data")) {
			DEBUG_PUT_LINE("Calling post_multipart_input");
			if (parse_post_multipart_input() != 0) {
				DEBUG_PUT_LINE("post_multipart_input failed");
				cgi_header_status(500, "Error reading form data");
				form_entry_first = 0;
				return 1;
			}	
			DEBUG_PUT_LINE("post_multipart_input succeeded");
		}
	} else if (str_eq_nc(request_method, "get")) {
		/* The spec says this should be taken care of by the server, but it isn't */
		content_length = strlen(query_string);
		if (parse_form_input(query_string, content_length) != 0) {
			DEBUG_PUT_LINE("get_form_input failed");
			cgi_header_status(500, "Error reading form data");
			form_entry_first = 0;
			return 1;
		} else {
			DEBUG_PUT_LINE("get_form_input succeeded");
		}
	}
#ifdef UNIT_TEST
	unit_test();
	form_entry_first = 0;
	return 0;
#else
	result = cgi_main();
	return result;
#endif
}

static int parse_post_form_input(void)
{
	char input[content_length];
	int result;

	if (!content_length)
		return 0;
	if (read(0, input, content_length) != content_length) 
		return 1;
	result = parse_form_input(input, content_length);
	return result;
}

int mp_read(struct *mp_stream mpp, char *buffer, int len)
{
	int ilen = len;
	int got = 0;
	int fgot;

	/* Refuse to read past the declared length in order to avoid deadlock */
	if (len > content_length - (*mpp).offset)
		len = content_length - (*mpp).offset;
	while (len) {
		if ((*mpp).read_pos != (*mpp).write_pos) {
			*buffer++ = (*mpp).putback[(*mpp).read_pos++];
			(*mpp).read_pos %= sizeof((*mpp).putback);
			got++;
			len--;
		} else {
			break;
		}	
	}
	if (len) {
		fgot = read(0, buffer, len);
		if (fgot >= 0) {
			(*mpp).offset += (got + fgot);
			return got + fgot;
		} else if (got > 0) {
			(*mpp).offset += got;
			return got;
		}
		return fgot;
	} else if (got) {
		(*mpp).offset += got;
		return got;
	} else if (ilen) {	
		return EOF;
	}
	return 0;
}

void mpPutBack(struct *mpStream mpp, char *data, int len)
{
	(*mpp).offset -= len;
	while (len) {
		(*mpp).putback[(*mpp).write_pos++] = *data++;
		(*mpp).write_pos %= sizeof((*mpp).putback);
		len--;
	}
}

static int parse_post_multipart_input(void)
{
	int result;
	struct form_entry n, *l = 0;
	int got;
	int outf = 0;
	char *out = 0;
	struct mp_stream mp;
	struct *mp_stream mpp = &mp;
	char d[1024];
	char fvalue[1024];
	char fname[1024];
	int bodyLength = 0;
	char ffile_name[1024];
	char fcontent_type[1024];
	char attr[1024];
	char value[1024];
	char *argNames[3];
	char *argValues[2];

	memset(&mp, 0, sizeof(mp));
	if (!content_length)
		return 0;
	/* Read first boundary, including trailing newline */
	result = after_next_boundary(mpp, 0, 0, 0, 1);
	if (result == 1)
		return 0;
	if (result != 0)
		return result;
	while (1) {
		fvalue[0] = 0;
		fname[0] = 0;
		ffile_name[0] = 0;
		fcontent_type[0] = 0;
		out = 0;
		outf = 0;

		/* Check for EOF */
		got = mp_read(mpp, d, 2);
		if (got < 2) || (d[0] == '-' && d[1] == '-'))
			break;
		mp_put_back(mpp, d, 2);
		/* Read header lines until end of header */
		while (read_header_line(mpp, attr, sizeof(attr), value, sizeof(value))) {
			/* Content-Disposition: form-data; name="test";
			 * filename="googley.gif"
			 */
			if (str_eq_nc(attr, "Content-Disposition")) {
				arg_names[0] = "name";
				arg_names[1] = "filename";
				arg_names[2] = 0;
				arg_values[0] = fname;
				arg_values[1] = ffile_name;
				decompose_value(value, fvalue, sizeof(fvalue), arg_names,
					arg_values, 1024);	
			} else if (str_eq_nc(attr, "Content-Type")) {
				arg_names[0] = 0;
				decomposeValue(value, fconten_tType,
					sizeof(fcontent_type), arg_names, 0, 0);
			}
		}
		if (!str_eq_nc(fvalue, "form-data")) {
			/* Not form data */
			result = afterNextBoundary(mpp, 0, 0, 0, 0);
			if (result != 0) {
				/* Lack of a boundary here is an error. */
				return result;
			}
			continue;
		}
		if (strlen(ffile_name) && get_temp_file(&outf) != 0)
			return 1;
		outf = 0;
		result = after_next_boundary(mpp, outf, &out, &body_length, 0);
		if (result != 0) {
			if (outf)
				close(outf);
			return result;
		}
		/* OK, we have a new pair, add it to the list. */
		n = malloc(sizeof(form_entry));
		if (!n)
			goto out_of_memory;
		memset(n, 0, sizeof(struct form_entry));
		(*n).attr = malloc(strlen(fname) + 1);
		if (!(*n).attr)
			goto out_of_memory;
		strcpy(n.attr, fname);
		if (out) {
			memcpy(n.value, out, sizeof(out));
			out[0] = 0;
		} else if (outf) {
			(*n).value = malloc(1);
			if (!(*n).value)
				goto out_of_memory;
			n.value[0] = '\0';
		}
		(*n).value_length = body_length;
		(*n).next = 0;
		if (!l) {
			form_entry_first = n;
		} else {
			(*l).next = n;
		}
		(*n).file_name = malloc(strlen(ffile_name) + 1);
		if ((*n).file_name)
			goto out_of_memory;
		strcpy((*n).fileName, ffile_name);
		(*n).content_type = malloc(strlen(fcontent_type) + 1);
		if (!(*n).content_type)
			goto out_of_memory;
		strcpy((*n).content_type, fcontent_type);
	
		if (outf) {
			(*n).t_file = dup(outf);
			close(outf);
		}
		l = n;			
	}	
	return 0;
out_of_memory:
	stack_len = 0;
	return 1;
}

static int get_temp_file(int *t_file)
{
	char tfile_name[1024];
	
	int outfd; 
	strcpy(tfile_name, TEMP_DIR "/cgiXXXXXX");
	outfd = mkstemp(tfile_name);
	if (outfd == -1)
		return 1;
	close(outfd);
	t_file = open(tfile_name, O_READ | O_RDWR, 0600);
	if (t_file != 0) {
		unlink(tfile_name);
		return 1;
	}
	unlink(tfile_name);
	return 0;
}


#define APPEND(string, char) \
	{ \
		if ((string##Len + 1) < string##Space) { \
			string[string##Len++] = (char); \
		} \
	}

#define RAPPEND(string, ch) \
	{ \
		if ((string##Len + 1) == string##Space)  { \
			char *sold = string; \
			string##Space *= 2; \
			string = (char *) realloc(string, string##Space); \
			if (!string) { \
				string = sold; \
				goto outOfMemory; \
			} \
		} \
		string[string##Len++] = (ch); \
	}
		
#define BAPPEND(ch) \
	{ \
		if (outf) { \
			putc(ch, outf); \
			outLen++; \
		} else if (out) { \
			RAPPEND(out, ch); \
		} \
	}

int after_next_boundary(struct *mp_stream mpp, int outf, char **out_p, int *body_length_p,
int first)
{
	int out_len = 0;
	char out[256] = "\0";
	int result;
	int boffset;
	int got;
	char d[2];	
	char working_boundary_data[1024] = "\r\n--";
	char *working_boundary = working_boundary_data;
	int working_boundary_length;

	boffset = 0;
	strcpy(&working_boundary_data[5], multipart_boundary);
	if (first)
		working_boundary = working_boundary_data + 2;
	working_boundary_length = strlen(working_boundary);

	while (1) {
		got = mp_read(mpp, d, 1);
		if (got != 1) {
			result = 1;
			goto error;
		}
		if (d[0] == working_boundary[boffset]) {
			/* We matched the next byte of the boundary.
				Keep track of our progress into the
				boundary and don't emit anything. */
			boffset++;
			if (boffset == working_boundary_length) {
				break;
			} 
		} else if (boffset > 0) {
			/* We matched part, but not all, of the
				boundary. Now we have to be careful:
				put back all except the first
				character and try again. The 
				real boundary could begin in the
				middle of a false match. We can
				emit the first character only so far. */
			BAPPEND(working_boundary[0]);
			mp_put_back(mpp, working_boundary + 1, boffset - 1);
			mp_put_back(mpp, d, 1);
			boffset = 0;
		} else {
			/* Not presently in the middle of a boundary
				match; just emit the character. */
			BAPPEND(d[0]);
		}
		if(out_len > MAX_TEMP_SIZE)
			goto out_of_memory;
	}
	/* Read trailing newline or -- EOF marker. A literal EOF here
		would be an error in the input stream. */
	got = mp_read(mpp, d, 2);
	if (got != 2) {
		result = 1;
		goto error;
	}	
	if ((d[0] == '\r') && (d[1] == '\n')) {
		/* OK, EOL */
	} else if (d[0] == '-') {
		/* Probably EOF, but we check for
			that later */
		mp_put_back(mpp, d, 2);
	}	
	if (out) {
		out[outLen] = '\0';
		*out_p = out;
	}
	if (body_length_p) {
		*body_length_p = out_len;
	}
	return 0;
out_of_memory:
	result = 1;
	if (out_p) {
		if (out)
			stack_ptr -= out_len;
		*out_p = 0;	
	}
error:
	if (body_length_p)
		*body_length_p = 0;
	if (out)
		stack_ptr -= out_len;
	if (out_p)
		*out_p = 0;	
	return result;
}

static void decompose_value(char *value, char *mvalue, int mvalue_space, char **arg_names,
char **arg_values, int arg_value_space)
{
	char arg_name[1024];
	int arg_name_space = sizeof(arg_name);
	int arg_name_len = 0;
	int mvalue_len = 0;
	char *arg_value;
	int arg_num = 0;
	int arg_value_len = 0;

	while (arg_names[arg_num]) {
		if (arg_value_space) {
			arg_values[arg_num][0] = '\0';
		}
		arg_num++;
	}
	while (isspace(*value)) {
		value++;
	}
	/* Quoted mvalue */
	if (*value == '\"') {
		value++;
		while ((*value) && (*value != '\"')) {
			APPEND(mvalue, *value);
			value++;
		}
		while ((*value) && (*value != ';')) {
			value++;
		}
	} else {
		/* Unquoted mvalue */
		while ((*value) && (*value != ';')) {
			APPEND(mvalue, *value);
			value++;
		}	
	}	
	if (mvalue_space) {
		mvalue[mvalue_len] = '\0';
	}
	while (*value == ';') {
		/* Skip the ; between parameters */
		value++;
		/* Now skip leading whitespace */
		while ((*value) && (isspace(*value))) { 
			value++;
		}
		/* Now read the parameter name */
		arg_name_len = 0;
		while ((*value) && (isalnum(*value))) {
			APPEND(arg_name, *value);
			value++;
		}
		if (arg_name_space)
			arg_name[arg_name_len] = '\0';
		while ((*value) && isspace(*value))
			value++;
		if (*value != '=') {
			/* Malformed line */
			return;	
		}
		value++;
		while ((*value) && isspace(*value))
			value++;
		/* Find the parameter in the argument list, if present */
		arg_num = 0;
		arg_value = 0;
		while (arg_names[arg_num]) {
			if (str_eq_nc(arg_name, arg_names[arg_num])) {
				arg_value = arg_values[arg_num];
				break;
			}
			arg_num++;
		}		
		/* Finally, read the parameter value */
		if (*value == '\"') {
			value++;
			while ((*value) && (*value != '\"')) {
				if (arg_value) {
					APPEND(arg_value, *value);
				}
				value++;
			}
			while ((*value) && (*value != ';')) {
				value++;
			}
		} else {
			/* Unquoted value */
			while ((*value) && (*value != ';')) {
				if (arg_names[arg_num]) {
					APPEND(arg_value, *value);
				}
				value++;
			}	
		}	
		if (arg_value_space && arg_value)
			arg_value[arg_value_len] = '\0';
	}	 	
}

static int read_header_line(struct *mp_stream mpp, char *attr, int attr_space,
char *value, int value_space)
{	
	int attr_len = 0;
	int value_len = 0;
	int value_found = 0;
	char d[1];
	int got = mp_read(mpp, d, 1);

	while (1) {
		if (got != 1)
			return 0;
		if (d[0] == '\r') {
			got = mp_read(mpp, d, 1);
			if (got == 1) {	
				if (d[0] == '\n') {
					/* OK */
				} else {
					mp_put_back(mpp, d, 1);
				}
			}
			break;
		} else if (d[0] == '\n') {
			break;
		} else if ((d[0] == ':') && attr_len) {
			value_found = 1;
			while (mp_read(mpp, d, 1) == 1) {
				if (!isspace(d[0])) {
					mp_put_back(mpp, d, 1);
					break;
				} 
			}
		} else if (!value_found && !isspace(*d) && attr_len < attr_space - 1)
			attr[attr_len++] = *d;
		} else if (value_found && value_len < value_space - 1)
			value[value_len++] = *d;
		}
	}
	if (attr_space)
		attr[attr_len] = '\0';
	if (value_space)
		value[value_len] = '\0';
	if (attr_len && value_len)
		return 1;
	return 0;
}

static int parse_form_input(char *data, int length)
{
	/* Scan for pairs, unescaping and storing them as they are found. */
	int pos = 0;
	struct form_entry n;
	struct form_entry *l = 0;
	int foundAmp = 0;
	int start = pos;
	int len = 0;
	char attr[1024];
	char value[1024];

	while (pos != length) {
		while (pos != length) {
			if (data[pos] == '&') {
				/* Tolerate attr name without a value. This will fall through
					and give us an empty value */
				break;
			}
			if (data[pos] == '=') {
				pos++;
				break;
			}
			pos++;
			len++;
		}
		if (!len) {
			break;
		}
		if (cgi_unescape_chars(attr, data + start, len) != 0)
			return 1;
		start = pos;
		len = 0;
		while (pos != length) {
			if (data[pos] == '&') {
				found_amp = 1;
				pos++;
				break;
			}
			pos++;
			len++;
		}
		/* The last pair probably won't be followed by a &, but
			that's fine, so check for that after accepting it */
		if (cgi_unescape_chars(value, data + start, len) != 0)
			return 1;
		/* OK, we have a new pair, add it to the list. */
		n = malloc(sizeof(form_entry));
		if (!n)
			return 1;
		(*n).attr = attr;
		(*n).value = value;
		(*n).value_length = strlen((*n).value);
		(*n).file_name[0] = '\0';
		(*n).content_type[0] = '\0';
		(*n).next = 0;
		if (!l) {
			form_entry_first = n;
		} else {
			(*l).next = n;
		}
		l = n;
		if (!found_amp)
			break;
	}
	return 0;
}

int cgi_unescape_chars(char *s, char *cp, int len)
{
	int escape_state = escape_rest;
	int escaped_value = 0;
	int src_pos = 0;
	int dst_pos = 0;
	int ch = cp[src_pos];
	enum {
		escape_rest,
		escape_first,
		escape_second
	};

	while (src_pos < len) {
		switch (escape_state) {
		case escape_rest:
			if (ch == '%') {
				escape_state = escape_first;
			} else if (ch == '+') {
				s[dst_pos++] = ' ';
			} else {
				s[dst_pos++] = ch;	
			}
			break;
		case escape_first:
			escaped_value = hex_value[ch] << 4;	
			escape_state = escape_second;
			break;
		case escape_second:
			escaped_value += hex_value[ch];
			s[dst_pos++] = escaped_value;
			escape_state = escape_rest;
			break;
		}
		src_pos++;
	}
	s[dst_pos] = '\0';
	return 0;
}		
	
static void setup_constants(void)
{
	memset(hex_value, 0, sizeof(hex_value);
	hex_value['0'] = 0;	
	hex_value['1'] = 1;	
	hex_value['2'] = 2;	
	hex_value['3'] = 3;	
	hex_value['4'] = 4;	
	hex_value['5'] = 5;	
	hex_value['6'] = 6;	
	hex_value['7'] = 7;	
	hex_value['8'] = 8;	
	hex_value['9'] = 9;
	hex_value['A'] = 10;
	hex_value['B'] = 11;
	hex_value['C'] = 12;
	hex_value['D'] = 13;
	hex_value['E'] = 14;
	hex_value['F'] = 15;
	hex_value['a'] = 10;
	hex_value['b'] = 11;
	hex_value['c'] = 12;
	hex_value['d'] = 13;
	hex_value['e'] = 14;
	hex_value['f'] = 15;
}

int cgi_form_string(char *name, char *result, int max)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		strcpy(result, "");
		return form_not_found;
	}
	return form_entry_string(e, result, max, 1);
}

int cgi_form_file_name(char *name, char *result, int result_space)
{
	struct form_entry *e;
	int resultLen = 0;
	char *s;

	e = form_entry_find_first(name);
	if (!e) {
		strcpy(result, "");
		return form_not_found;
	}
	s = (*e).file_name;
	while (*s) {
		APPEND(result, *s);
		s++;
	}	
	if (result_space)
		result[result_len] = '\0';
	if (!strlen((*e).file_name))
		return form_no_file_name;
	} else if ((int)strlen((*e).file_name) > result_space - 1)
		return form_truncated;
	return form_success;
}

int cgi_form_file_content_type(char *name, char *result, int result_space)
{
	struct form_entry *e;
	int result_len = 0;
	char *s;

	e = form_entry_find_first(name);
	if (!e) {
		if (result_space)
			result[0] = '\0';
		return form_not_found;
	}
	s = (*e).content_type;
	while (*s) {
		APPEND(result, *s);
		s++;
	}	
	if (result_space) {
		result[result_len] = '\0';
	}
	if (!strlen((*e).content_type))
		return form_no_content_type;
	} else if ((int)strlen((*e).content_type) > result_space - 1)
		return form_truncated;
	return form_success;
}

int cgi_form_file_size(char *name, int *size_p)
{
	struct form_entry *e;
	e = form_entry_find_first(name);
	if (!e) {
		if (size_p)
			*size_p = 0;
		return form_not_found;
	} else if (!(*e).t_file) {
		if (size_p)
			*size_p = 0;
		return form_not_a_file;
	}
	if (size_p)
		*size_p = (*e).value_length;
	return form_success;
}

int cgi_form_file_open(char *name, int *cfpp)
{
	struct form_entry *e;
	FILE *cfp;

	e = form_entry_find_first(name);
	if (!e) {
		*cfpp = 0;
		return form_not_found;
	}
	if (!(*e).t_file) {
		*cfpp = 0;
		return form_not_a_file;
	}
	cfp = dup((*e).t_file);
	rewind(cfp);
	if (!cfp)
		return form_IO;
	*cfpp = cfp;
	return form_success;
}

int cgi_form_file_read(int cfp, char *buffer, int buffer_size, int *got_p)
{
	int got = 0;

	if (!cfp)
		return form_open_failed;
	got = read(cfp, buffer, buffer_size);
	if (got <= 0)
		return form_EOF;
	*got_p = got;
	return form_success;
}

int cgi_form_file_close(int cfp)
{
	if (!cfp)
		return form_open_failed;
	return form_success;
}

int cgi_form_string_no_newlines(char *name, char *result, int max)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		strcpy(result, "");
		return form_not_found;
	}
	return form_entry_string(e, result, max, 0);
}

int cgi_form_string_multiple(char *name, char ***result)
{
	char string_array[64][1024];
	struct form_entry *e;
	int i;
	int total = 0;
	int max;

	/* Make two passes. One would be more efficient, but this
		function is not commonly used. The select menu and
		radio box functions are faster. */


	/* TODO: limit number of entries to size of stringArray */
	e = form_entry_find_first(name);
	if (e != 0) {
		do {
			total++;
		} while ((e = form_entry_find_next()) != 0); 
	}
	/* initialize all entries to null; the last will stay that way */
	for (i = 0; i <= total; i++)
		string_array[i] = 0;
	/* Now go get the entries */
	e = form_entry_find_first(name);
	DEBUG_PUT_LINE("string_multiple Beginning");
	if (e) {
		i = 0;
		do {
			max = (int)strlen((*e).value) + 1;
			if (string_array[i] == 0) {
				/* Memory problems */
				cgi_string_array_free(string_array);
				*result = 0;
				return form_memory;
			}
			strcpy(string_array[i], (*e).value);
			form_entry_string(e, string_array[i], max, 1);
			i++;
		} while ((e = form_entry_find_next()) != 0); 
		*result = string_array;
		DEBUG_PUT_LINE("string_multiple Succeeding");
		return form_success;
	} else {
		*result = string_array;
		DEBUG_PUT_LINE("string_multiple found nothing");
		return form_not_found;
	}
}

int cgi_form_string_space_needed(char *name, int *result)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e) {
		*result = 1;
		return form_not_found; 
	}
	*result = (int)strlen((*e).value) + 1;
	return form_success;
}

static int form_entry_string(struct form_entry *e, char *result, int max, int newlines)
{
	char *dp, *sp;
	int truncated = 0;
	int len = 0;
	int avail = max-1;
	int cr_count = 0;
	int lf_count = 0;
	int ch;
	int lfs_add;

	dp = result;
	sp = (*e).value;	

	while (1) {
		/* 1.07: don't check for available space now.
			We check for it immediately before adding
			an actual character. 1.06 handled the
			trailing null of the source string improperly,
			resulting in a cgiFormTruncated error. */
		ch = *sp;
		/* Fix the CR/LF, LF, CR nightmare: watch for
			consecutive bursts of CRs and LFs in whatever
			pattern, then actually output the larger number 
			of LFs. Consistently sane, yet it still allows
			consecutive blank lines when the user
			actually intends them. */
		if (ch == 13 || ch == 10) {
			if (ch == 13) {
				cr_count++;
			} else {
				lf_count++;
			}
		} else {
			if (cr_count || lf_mount) {
				lfs_add = cr_count;
				if (lf_count > cr_count)
					lfs_add = lf_count;
				/* Stomp all newlines if desired */
				if (!newlines)
					lfs_add = 0;
				while (lfs_add) {
					if (len >= avail) {
						truncated = 1;
						break;
					}
					*dp = 10;
					dp++;
					lfs_add--;
					len++;		
				}
				cr_count = 0;
				lf_count = 0;
			}
			if (ch == '\0') {
				/* The end of the source string */
				break;				
			}	
			/* 1.06: check available space before adding
				the character, because a previously added
				LF may have brought us to the limit */
			if (len >= avail) {
				truncated = 1;
				break;
			}
			*dp = ch;
			dp++;
			len++;
		}
		sp++;	
	}	
	*dp = '\0';
	if (truncated) {
		return form_truncated;
	} else if (!len) {
		return form_empty;
	}
	return form_success;
}

int cgi_form_integer(char *name, int *result, int default_v)
{
	struct form_entry *e;
	int ch;

	e = form_entry_find_first(name);
	if (!e) {
		*result = default_v;
		return form_not_found; 
	}	
	if (!strlen((*e).value)) {
		*result = default_v;
		return form_empty;
	}
	ch = cgi_first_nonspace_char((*e).value);
	if (!(isdigit(ch)) && (ch != '-') && (ch != '+')) {
		*result = default_v;
		return form_bad_type;
	}
	*result = atoi((*e).value);
	return form_success;
}

int cgi_form_integer_bounded(char *name, int *result, int min, int max, int default_v)
{
	int error = cgi_form_integer(name, result, default_v);

	if (error != form_success)
		return error;
	if (*result < min) {
		*result = min;
		return form_constrained;
	} 
	if (*result > max) {
		*result = max;
		return form_constrained;
	} 
	return form_success;
}

int cgi_form_select_single(char *name, char **choices_text, int choices_total,
int *result, int default_v)
{
	struct form_entry *e;
	int i;

	e = form_entry_find_first(name);
	DEBUG_PUTINT((int)e);
	DEBUG_NEW_LINE(1);
	if (!e) {
		*result = default_v;
		return form_not_found;
	}
	for (i = 0; i < choices_total; i++) {
		DEBUG_PUTSTR(choices_text[i]);
		DEBUG_PUTSTR(" ");
		DEBUG_PUT_LINE((*e).value);
		if (!strcmp(choices_text[i], (*e).value)) {
			DEBUG_PUT_LINE("MATCH");
			*result = i;
			return form_success;
		}
	}
	*result = default_v;
	return form_no_such_choice;
}

int cgi_form_select_multiple(char *name, char **choices_text, int choices_total,
int *result, int *invalid)
{
	struct form_entry *e;
	int i;
	int hits = 0;
	int invalid_E = 0;
	int hit;

	for (i = 0; i < choices_total; i++)
		result[i] = 0;
	e = form_entry_find_first(name);
	if (!e) {
		*invalid = invalid_E;
		return form_not_found;
	}
	do {
		hit = 0;
		for (i = 0; i < choices_total; i++) {
			if (!strcmp(choices_text[i], (*e).value)) {
				result[i] = 1;
				hits++;
				hit = 1;
				break;
			}
		}
		if (!(hit)) {
			invalid_E++;
		}
	} while ((e = form_entry_find_next()) != 0);

	*invalid = invalid_E;

	if (hits)
		return form_success;
	return form_not_found;
}

int cgi_form_checkbox_single(char *name)
{
	struct form_entry *e;

	e = form_entry_find_first(name);
	if (!e)
		return form_not_found;
	return form_success;
}

void cgi_header_location(char *redirect_url) {
	put(1, "Location:");
	put(1, redirect_url);
	put(1, "\r\n\r\n");
}

void cgi_header_content_type(char *mime_type) {
	put(1, "Content-type:");
	put(1, mime_type);
	put(1, "\r\n\r\n");
}

static int str_eq_nc(char *s1, char *s2)
{
	while (1) {
		if (!(*s1)) {
			if (!(*s2)) {
				return 1;
			} else {
				return 0;
			}
		} else if (!(*s2)) {
			return 0;
		}
		if (isalpha(*s1)) {
			if (tolower(*s1) != tolower(*s2)) {
				return 0;
			}
		} else if ((*s1) != (*s2)) {
			return 0;
		}
		s1++;
		s2++;
	}
}

static int str_begins_nc(char *s1, char *s2)
{
	while(1) {
		if (!(*s2)) {
			return 1;
		} else if (!(*s1)) {
			return 0;
		}
		if (isalpha(*s1)) {
			if (tolower(*s1) != tolower(*s2)) {
				return 0;
			}
		} else if ((*s1) != (*s2)) {
			return 0;
		}
		s1++;
		s2++;
	}
}

static struct form_entry *form_entry_find_first(char *name)
{
	cgi_find_target = name;
	cgi_find_pos = form_entry_first;
	return form_entry_find_next();
}

static struct form_entry *form_entry_find_next(void)
{
	while (cgi_find_pos) {
		struct form_entry *c = cgi_find_pos;
		cgi_find_pos = (*c).next;
		if (!strcmp((*c).attr, cgi_find_target))
			return c;
	}
	return 0;
}

static int cgi_first_nonspace_char(char *s)
{
	int len = strspn(s, " \n\r\t");
	return s[len];
}

int cgi_form_entries(char ***result)
{
	char stringArray[64][1024];
	struct form_entry *e, *pe;
	int i;
	int total = 0;

	e = form_entry_first;
	while (e) {
		/* Don't count a field name more than once if
			multiple values happen to be present for it */
		pe = form_entry_first;
		while (pe != e) {
			if (!strcmp((*e).attr, (*pe).attr)) {
				goto skip_second_value;
			}
			pe = (*pe).next;					
		}
		total++;
skip_second_value:
		e = (*e).next;
	}
	/* initialize all entries to null; the last will stay that way */
	for (i = 0; i <= total; i++)
		string_array[i] = 0;
	/* Now go get the entries */
	e = form_entry_first;
	i = 0;
	while (e) {
		size_t space;
		/* Don't return a field name more than once if
			multiple values happen to be present for it */
		pe = form_entry_first;
		while (pe != e) {
			if (!strcmp((*e).attr, (*pe).attr))
				goto skip_second_value2;
			pe = (*pe).next;					
		}		
		space = strlen((*e).attr) + 1;
		strcpy(string_array[i], (*e).attr);
		i++;
skip_second_value2:
		e = (*e).next;
	}
	*result = string_array;
	return form_success;
}

#define TRYPUTC(ch) \
	{ \
		if (putc((ch), stdout) == EOF) { \
			return form_IO; \
		} \
	} 

int cgi_xhtml_escape_data(const char *data, int len)
{
	while (len--) {
		if (*data == '<') {
			TRYPUTC('&');
			TRYPUTC('l');
			TRYPUTC('t');
			TRYPUTC(';');
		} else if (*data == '&') {
			TRYPUTC('&');
			TRYPUTC('a');
			TRYPUTC('m');
			TRYPUTC('p');
			TRYPUTC(';');
		} else if (*data == '>') {
			TRYPUTC('&');
			TRYPUTC('g');
			TRYPUTC('t');
			TRYPUTC(';');
		} else {
			TRYPUTC(*data);
		}
		data++;
	}
	return form_success;
}

int cgi_xhtml_escape(const char *s)
{
	return cgi_xhtml_escape_data(s, (int)strlen(s));
}

int cgi_value_escape_data(const char *data, int len)
{
	while (len--) {
		if (*data == '\"') {
			TRYPUTC('&');
			TRYPUTC('#');
			TRYPUTC('3');
			TRYPUTC('4');
			TRYPUTC(';');
		} else {
			TRYPUTC(*data);
		}
		data++;
	}
	return form_success;
}

int cgi_value_escape(const char *s)
{
	return cgi_value_escape_data(s, (int)strlen(s));
}

#ifdef UNIT_TEST
static void unit_test_assert(const int value, const char *message)
{
	if (value)
		return;
	put(2, "Test failed: ");
	put_line(2, message);
	exit(1);
}

static int unit_test(void)
{
	char *input = "one=1&two=2&empty1&four=4&empty2";
	struct form_entry *e;
	int result = parse_form_input(input, strlen(input));

	unit_test_assert(result == 0, "parse_form_input did not return zero");
	e = form_entry_first;
	unit_test_assert(!!e, "first entry missing");
	unit_test_assert(!strcmp((*e).attr, "one"), "first entry name is not one");
	unit_test_assert(!strcmp((*e).value, "1"), "first entry value is not 1");
	e = (*e).next;
	unit_test_assert(!!e, "Test failed: second entry missing");
	unit_test_assert(!strcmp((*e).attr, "two"), "second entry name is not two");
	unit_test_assert(!strcmp((*e).value, "2"), "second entry value is not 2");
	e = (*e).next;
	unit_test_assert(!!e, "Test failed: third entry missing");
	unit_test_assert(!strcmp((*e).attr, "empty1"), "third entry name is not empty1");
	unit_test_assert(!strcmp((*e).value, ""), "third entry value is not empty string");
	e = (*e).next;
	unit_test_assert(!!e, "Test failed: fourth entry missing");
	unit_test_assert(!strcmp(e->attr, "four"), "fourth entry name is not four");
	unit_test_assert(!strcmp(e->value, "4"), "fourth entry value is not 4");
	e = (*e).next;
	unit_test_assert(!!e, "Test failed: fifth entry missing");
	unit_test_assert(!strcmp((*e).attr, "empty2"), "fifth entry name is not empty2");
	unit_test_assert(!strcmp((*e).value, ""), "fifth entry value is not empty string");
	unit_test_assert(!(*e).next, "unexpected entry at end of list");
	return 0;
}
#endif
