#!/bin/sh
CFLAGS="-nostdinc -fno-builtin -fno-unwind-tables -fno-stack-protector \
-std=c89 -Wpedantic -Wall -Werror $CFLAGS"
AFLAGS="--warn --fatal-warnings $AFLAGS"

: ${CC:=gcc}
: ${AS:=as}
: ${AR:=ar}
: ${PREFIX:=/usr/local}
: ${ARCH:=$(uname -m)}

run() {
  printf '%s\n' "$*"
  "$@"
}
 
install() {
  [ -f libfndn.a ] || build

  run mkdir -p "${PREFIX}/lib"
  run mkdir -p "${PREFIX}/share/doc/libfndn"
  run cp -f CHANGELOG README "${PREFIX}/share/doc/libfndn/"
  run cp -f libfndn.a fndn-gcc.specs "${PREFIX}/lib/"
  run cp -af include "$PREFIX"
  run cp -af fndn-gcc "${PREFIX}/bin/"
}

uninstall() {
  run rm -fr "${PREFIX}/share/doc/libfndn" "${PREFIX}/include"
  run rm -f "${PREFIX}/lib/libfndn.a" "${PREFIX}/lib/fndn-gcc.specs"
}

build() {
  if [ "$ARCH" != "arm" ] && [ "$ARCH" != "i386" ]; then
    echo "Unsupported architecture: $ARCH" >&2
    exit 1
  fi
  run chdir source
  run "$AS" $AFLAGS -o crt-${ARCH}.o crt-${ARCH}.s
  for source in $(ls -1 */${ARCH}/*.s | cut -d'.' -f1); do
    run "$AS" $AFLAGS -o ${source}.o ${source}.s
  done
  for source in $(ls -1 */*.c | cut -d'.' -f1); do
    run "$CC" $CFLAGS -I../include -c -o ${source}.o ${source}.c
  done
  run chdir ..
  run "$AR" rcs libfndn.a source/crt-${ARCH}.o source/*/${ARCH}/*.o source/*/*.o
}

clean() {
  run rm -f source/crt-${ARCH}.o source/*/${ARCH}/*.o source/*/*.o libfndn.a
}

if [ "$#" -gt 0 ]; then
  if ! type $1 | grep -q "function$"; then
    echo "$(basename $0): Unknown verb: $1" >&2
    exit 1
  else
    "$@"
  fi
else
  build
fi
